
// BlastsManager.java
// Andrew Davison, May 2006, ad@fivedots.coe.psu.ac.th

/* BlastsManager manages NUM_BLASTS BlastSprites.
   When MWPanel calls requestBlast() it starts one
   of the inactive sprites moving (if one is available). 
   The sprite starts at the current wand head position, and 
   moves along the current direction of the wand, by 
   STEP_SIZE pixels in each update.

   BlastsManager updates and draws all of its 
   sprites when MWPanel calls its updateSprites() 
   and drawSprites() methods.

   BlastSprite ia a subclass of the Sprite class, which
   is described in KGPJ chapter 11.
*/

import java.awt.*;


public class BlastsManager
{
  private static final int NUM_BLASTS = 10;
  private static final double STEP_SIZE = 10.0;   
    // step size taken by a blast in each update

  private BlastSprite sprites[];



  public BlastsManager(int width, int height, ImagesLoader imsLd, int period)
  {
    // create NUM_BLASTS inactive sprites, so they're ready for use
    sprites = new BlastSprite[NUM_BLASTS];
    for (int i=0; i < NUM_BLASTS; i++)
      sprites[i] = new BlastSprite(width, height, imsLd, period);
    
  }  // end of BlastsManager()


  public void updateSprites()
  {
    for (int i=0; i < NUM_BLASTS; i++)
      sprites[i].updateSprite();
  }  // end of updateSprites()


  public void drawSprites(Graphics g)
  {
    for (int i=0; i < NUM_BLASTS; i++)
      sprites[i].drawSprite(g);
  }  // end of drawSprites()


  public void requestBlast(int xHead, int yHead, double lineAngle)
  /* Called from MWPanel to fire a blast from
     the head of the wand. However, a blast is only emitted if
     there's an inactive BlastSprite currently available.
  */
  {
    // calculate sprite's steps in the x- and y- directions
    int xStep = (int) Math.round( -STEP_SIZE * Math.cos(lineAngle) );
    int yStep = (int) Math.round( -STEP_SIZE * Math.sin(lineAngle) );
      /* the negative step makes the direction opposite to the 
         direction defined by the wand head-to-tail */

    // fire an inactive blast sprite (if there's one available)
    for (int i=0; i < NUM_BLASTS; i++)
      if (!sprites[i].isActive()) {
        sprites[i].startMoving(xHead, yHead, xStep, yStep);
/*
        System.out.println("Sprite " + i + " started at (" + 
                              xHead + "," + yHead + "); step: (" + 
                              xStep + "," + yStep + ")");
*/
        break;
      }
  }  // end of requestBlast()

}  // end of BlastsManager class

