
// BlobsLine.java
// Andrew Davison, May 2006, ad@fivedots.coe.psu.ac.th

/* Two BlobManagers manage blobs for two different colours
   When all the pixels in the current image have been assigned to 
   blobs (or discarded), a line is calculated between the two largest
   blobs of the two colours. One blob is centered on the head of the 
   line, the other on it's tail.

   The coordinates of the line's head, and its angle to the horizontal 
   are retrieved by ImageAnalyzer calling getLineHead() and getAngle().

   Our assumption is that the line corresponds to the user's wand.
*/

import java.awt.*;
import java.util.*;
import java.text.DecimalFormat;


public class BlobsLine
{
  // IDs for the possible line colours
  public static final int UNMATCHED_COLOUR = 0;   // used by ImageAnalyzer
  private static final int HEAD_COLOUR = 1;
  private static final int TAIL_COLOUR = 2;

  // blob managers for the head and tail colours of the line
  private BlobsManager blobManHead, blobManTail;  

  // line-related data
  private Point lineHead = null;
  private double lineAngle = 0;

  private DecimalFormat df = new DecimalFormat("0.#");  // 1 dp



  public BlobsLine(Color headColour, Color tailColour)
  {
    blobManHead = new BlobsManager(headColour);
    blobManTail = new BlobsManager(tailColour);
  }  // end of BlobsLine()


  public void clear()
  // delete all the blobs in the two managers
  {  blobManHead.clear();  
     blobManTail.clear();
  }


  public int matchesColour(int pixel)
  /* is the (r,g,b) pixel colour close to one of the blob colours
     for the line? Return the relevant line colour ID, 
     UNMATCHED_COLOUR otherwise. */
  {
    if (blobManHead.matchesColour(pixel))   // pixel close to line head colour
      return HEAD_COLOUR;
    else if (blobManTail.matchesColour(pixel))   // close to tail colour
      return TAIL_COLOUR;
    else
      return UNMATCHED_COLOUR;
  }  // end of matchesColour()



  public void addPoint(int x, int y, int colourID)
  /* Pass (x,y) to the blobs managers for the relevant 
     colour ID. */
  {  
    if (colourID == HEAD_COLOUR)
      blobManHead.addPoint(x,y);
    else if (colourID == TAIL_COLOUR)
      blobManTail.addPoint(x,y);
    else
      System.out.println("Could not add point for colour number " + colourID);
  }  // end of addPoint()


 
  public void calculateLine()
  /*  Find the largest blobs of the two colours, and calculate
      the line's head pt, and its angle to the horizontal.
  */
  {
    // get the biggest blob for the head colour 
    int headBlobIdx = blobManHead.findLargest();  
    if (headBlobIdx == -1)   // no blob found
      return;   // don't change head or angle values

    // get the biggest blob for the tail colour 
    int tailBlobIdx = blobManTail.findLargest(); 
    if (tailBlobIdx == -1)   // no blob found
      return;    // don't change head or angle values

    // found two largest blobs; so get their centers
    lineHead = blobManHead.getBlobCenter( headBlobIdx );
    Point tailPt = blobManTail.getBlobCenter( tailBlobIdx );
 /*
    System.out.println("Line Coords: hd=(" + df.format(lineHead.x) + "," + 
                                   df.format(lineHead.y) + ") -- tl=(" +
                                   df.format(tailPt.x) + "," + 
                                   df.format(tailPt.y) + ")" );
 */

    // get angle of line head to the horizontal
    int xDist = tailPt.x - lineHead.x;
    int yDist = tailPt.y - lineHead.y;
    if (xDist == 0)  
      xDist = 1;  // avoid division by zero

    // change lineAngle and line
    lineAngle = Math.atan( ((double)yDist)/xDist );
    // System.out.println("Angle: " + df.format(Math.toDegrees(lineAngle)) );

  }  // end of calculateLine()


  // -------- access line info (called by ImageAnalyzer) --------------

  public Point getLineHead()
  { return lineHead; }

  public double getAngle()
  { return lineAngle; }

}  // end of BlobsLine class