
// BlobsManager.java
// Andrew Davison, May 2006, ad@fivedots.coe.psu.ac.th

/* Manage all the blobs created for a given colour.

   addPoint() adds a pixel coordinate to a blob that is
   closest to its (x,y) coordinate.

   findLargest() returns the index of the largest blob
   (which must be bigger than MIN_POINTS).
   getBlobCenter() returns the center of a given blob.
*/

import java.awt.*;
import java.util.*;


public class BlobsManager
{
  // max distance*distance to the colour
  private static final int MAX_DIST2 = 5000;

  /* Minimum no. of points necessary for a blob to be considered
     when the manager is looking for a largest blob. */
  private static final int MIN_POINTS = 100;


  // RGB components for this manager's colours
  private int redCol, greenCol, blueCol;

  private ArrayList<Blob> blobs;   // of Blob objects
  private int currBlobIdx;     // index of last blob that was updated
  

  public BlobsManager(Color c)
  {
    redCol = c.getRed();
    greenCol = c.getGreen();
    blueCol = c.getBlue();

    blobs = new ArrayList<Blob>();
    currBlobIdx = -1;
  }  // end of BlobsManager()


  public void clear()
  // delete all the blobs
  {  blobs.clear();  
     currBlobIdx = -1;
  }


  public boolean matchesColour(int pixel)
  /* is the (r,g,b) pixel colour close to the manager's colour? */
  {
    // extract RGB components from the current pixel as integers
    int r = (pixel>>16)&255;
    int g = (pixel>>8)&255;
    int b = pixel&255;

    int redDiff = r - redCol;
    int greenDiff = g - greenCol;
    int blueDiff = b - blueCol;

    return (((redDiff * redDiff) + 
             (greenDiff * greenDiff) + 
             (blueDiff * blueDiff)) < MAX_DIST2);
  }  // end of matchesColour()



  public void addPoint(int x, int y)
  /* Add (x,y) to an existing blob if close enough, or create
     a new blob for it.
  */
  {  
    int blobIndex = findCloseBlob(x,y);
    if (blobIndex != -1) {   // found a blob close to (x,y)
      Blob b = (Blob) blobs.get(blobIndex);
      b.addPoint(x,y);
      currBlobIdx = blobIndex;
    }
    else {   // no close blob, so create a new one
      Blob b = new Blob();
      b.addPoint(x,y);
      blobs.add(b);
      currBlobIdx = blobs.size() - 1;
    }
  }  // end of addPoint()


  private int findCloseBlob(int x, int y)
  /* Find a blob that's close to (x,y), and return its
     index number. If no suitable blob is found then 
     return -1.
  */
  {
    Blob blob;

    // try current blob first
    if (currBlobIdx != -1) {
      blob = (Blob) blobs.get(currBlobIdx);
      if (blob.isClose(x,y))
        return currBlobIdx;
    }

    // otherwise try the others
    for(int i=0; i < blobs.size(); i++) { 
      if (i != currBlobIdx) {
        blob = blobs.get(i);
        if (blob.isClose(x,y))
          return i;
      }
    }
    return -1;   // didn't find a close blob
  }  // end of findCloseBlob()
 

  public int findLargest()
  /* Search through the blobs, and return the index
     of the largest, or -1 if none is found */
  {
    int largestIdx = -1;
    int maxPts = MIN_POINTS;    // blob must be at least this big

    int numPts;
    Blob blob;
    for(int i=0; i < blobs.size(); i++) { 
      blob = blobs.get(i);
      numPts = blob.numPoints();  // use blob's no. of points as size measure
      if (numPts > maxPts) {
        maxPts = numPts;
        largestIdx = i;
      }
    }
    return largestIdx;
  }  // end of findLargest()



  public Point getBlobCenter(int idx)
  // return the center point for the blob at position idx
  {
    Point p = null;
    if ((idx < 0) || (idx >= blobs.size()))
      System.out.println("No blob with that index: " + idx);
    else {
      Blob blob = blobs.get(idx);
      p = blob.getCenter();
    }
    return p;
  }  // end of getBlobCenter()


}  // end of BlobsManager class