
// ImageAnalyzer.java
// Andrew Davison, May 2006, ad@fivedots.coe.psu.ac.th

/* A thread is started which uses a JMFCapture object to take 
   a picture every SNAP_INTERVAL ms. Each image is analyzed by
   analyzeImage(). 
 
   The webcam picture should include a white strip (wand) with a 
   blob of colour at its head and a different one at its tail. 
   The colours are defined in HEAD_COLOUR and TAIL_COLOUR.

   The image is processed as an ARGB pixel array. Each pixel is
   compared to the wand colours, and if the pixel 
   colour is near one then the pixel's coordinate is 
   added to a relevant blob. If a pixel isn't near one of the
   colours then that pixel is made transparent.

   After the image has been completely examined, the BlobsLine 
   object calculates the coordinates of a line between the two largest
   head and tail blobs, and the line's angle relative to the horizontal.
   Our assumption is that the line corresponds to the user's wand. 

   ImageAnalyzer grabs the line's head coordinates and the line's angle and 
   passes them to the drawing panel thread by calling MWPanel's updateCamInfo(). 
   The changed image is also sent over -- it will be mostly transparent 
   except for the blobs.
*/

import java.awt.*;
import java.awt.image.*;

public class ImageAnalyzer extends Thread
{
  private static final int SNAP_INTERVAL = 120;  // ms; the time between snaps
     /* The value for SNAP_INTERVAL is based on timings that show 
        that a snap takes about 30ms to be carried out, and the 
        subsequent analysis of the image requires about 90 ms.
     */

  /* head and tail colours for the wand -- you should modify these to suit 
     the camera's properties and environmental issues such as lighting 
     and the background. */
  private static final Color HEAD_COLOUR = new Color(200, 200, 85);  // yellow
  private static final Color TAIL_COLOUR = new Color(50, 150, 200);  // blue
  // alternative red colours:  Color(205, 16, 80), Color(175, 60, 65)

  private static final int CLEAR = 0xffffff;  
    // black pixel, which will be rendered transparently since the image is ARGB


  private JMFCapture camera;     // JMF webcam object
  private BufferedImage camImage = null;  // the current webcam image
  private int imSize;                     // size of captured image
  private MWPanel mwPanel;        
            /* panel for displaying the snapped (and modified) images;
               it also utilizes the wand coord and angle information */
  private BlobsLine blobsLine;   
     /* BlobsLine manages the creation of the blobs that are used to
        calculate the wand's coordinates and its angle */

  private boolean running;



  public ImageAnalyzer(MWPanel lp, int sz)
  {
    mwPanel = lp;
    imSize = sz;
    blobsLine = new BlobsLine(HEAD_COLOUR, TAIL_COLOUR);
  }  // end of ImageAnalyzer()


  public void run()
  /* Use a JMFCapture object to take a picture every SNAP_INTERVAL ms.
     The image is analyzed to assign pixels to blobs inside blobsLine.
     These blobs are used to calculate line head coord and angle info, 
     which is passed, along with the image, to the drawing panel.
     The assumption is that the line is the user's wand.
  */
  {
    camera = new JMFCapture(imSize);   // initialize the webcam
    System.out.println("**Camera Ready**");

    long duration;
    BufferedImage im = null;
    running = true;

    while (running) {
	  long startTime = System.currentTimeMillis();
      // System.out.println("Start loading image");
      im = camera.grabImage();  // take a snap
      if (im == null) {
        System.out.println("Problem loading image");
        duration = System.currentTimeMillis() - startTime;
      }
      else {
        analyzeImage(im);
        blobsLine.calculateLine();     // calculate line coords and angle

        duration = System.currentTimeMillis() - startTime;
        // System.out.println("Duration: " + duration + " ms");

        // send the snapped image and line info to the JPanel for rendering
        mwPanel.updateCamInfo(im, blobsLine.getLineHead(),
                                  blobsLine.getAngle() );  
      }

      if (duration < SNAP_INTERVAL) {
        try {
          Thread.sleep(SNAP_INTERVAL-duration);  // wait until interval has passed
        } 
        catch (Exception ex) {}
      }
    }
    camera.close();    // close down the camera
  }  // end of run()


  // ----------------- image analysis ----------------------


  private void analyzeImage(BufferedImage im)
  /* The image is processed as a pixel array. Each pixel is compared to 
     the wand colours. If the pixel colour is near to one of them then 
     the pixel's coordinate is added to a blob for that colour. If the 
     pixel isn't close to a wand colour then the pixel is made transparent. */
  {
    if (im == null) {
      System.out.println("Input image is null");
      return;
    }

    blobsLine.clear();   // start with no blobs

    // extract pixels from the image into an array
    int imWidth = im.getWidth();
    int imHeight = im.getHeight();
    int [] pixels = new int[imWidth * imHeight];
    im.getRGB(0, 0, imWidth, imHeight, pixels, 0, imWidth);

    int xc = 0;    // hold the current pixel coordinates
    int yc = 0;

    // examine the pixels
    int i = 0;
    int colourID = BlobsLine.UNMATCHED_COLOUR;
    while(i < pixels.length) {
      colourID = blobsLine.matchesColour(pixels[i]); // test if matches a wand colour
      if (colourID != BlobsLine.UNMATCHED_COLOUR)  // it matches the head or tail colour
        blobsLine.addPoint(xc, yc, colourID);   // store the point with that colour
      else      // no match
        pixels[i] = CLEAR;   // make the pixel transparent

      // move to next coordinate in image
      xc++;
      if (xc == imWidth) {    // at end of row
        xc = 0;   // start next row
        yc++;
      }
      i++;
    }

    // update the image with the modified pixels data
    im.setRGB(0, 0, imWidth, imHeight, pixels, 0, imWidth);
  }  // end of analyzeImage()


  // --------------------- finish -----------------------

  public void closeDown()
  /* closeDown() blocks until run() has called JMFCapture.close() 
     and finished. This avoids nasty OS effects after the application 
     has terminated, but it means that clicking on the close box
     'freezes' the application for 1-2 secs.
  */
  { 
    running = false;
    while (!camera.isClosed()) {
      try {
        Thread.sleep(SNAP_INTERVAL);
      } 
      catch (Exception ex) {}
    }
  } // end of closeDown()

}  // end of ImageAnalyzer class