
// MWPanel.java
// Andrew Davison, May 2006, ad@fivedots.coe.psu.ac.th

/* The game's drawing surface.
   Active rendering is used to update the JPanel. See WormP for
   another example (KGPJ chapter 3), with additional 
   statistics generation. In this version, the timing is done with
   Java 5's System.nanoTime() method.

   As the user moves a 'wand' in front of a webcam, a representation
   of the wand is drawn on-screen, and 'blast' sprites are emitted
   from its head. The sprites move in a general right-to-left direction 
   dictated by the way the wand is currently pointing.

   The sprites disappear off the left side of the screen without doing 
   anything, but they could easily form part of a 2D shooting game.

   The webcam image is analyzed by a threaded ImageAnalyzer object,
   which periodically sends the coordinates for the wand head, the 
   wand's angle to the horizontal, and the webcam image. 

   The blast sprites are managed by a BlastsManager object

   Various images (the background, the wand image, the images for
   the sprites) are loaded using the ImagesLoader class which is
   described in KGPJ chapter 6.

   -----
   MWPanel and the ImageAnalyzer are both threaded, and share three
   data strutures: the webcam image, the wand's head point, and the
   wand's angle to the horizontal. Care must be taken with 
   shared data between threads. In particular, ImageAnalyzer may
   update them while MWPanel is using them in gameRender(),
   drawWand(), and fireWand(). Rather than add in synchronization
   code, which would slow things down, I've chosen not to worry about
   potential conflicts which, even if they occur, won't cause any serious
   problems. I discuss the reasoning behind this in more detail in the
   chapter.
*/

import javax.swing.*;
import java.awt.image.*;
import java.awt.event.*;
import java.awt.*;
import java.awt.geom.*;


public class MWPanel extends JPanel implements Runnable
{
  private static final int PWIDTH = 700;   // size of panel
  private static final int PHEIGHT = 256; 

  private static final int NO_DELAYS_PER_YIELD = 16;
  /* Number of frames with a delay of 0 ms before the animation thread yields
     to other running threads. */
  private static int MAX_FRAME_SKIPS = 5;
    // no. of frames that can be skipped in any one animation loop
    // i.e the games state is updated but not rendered

  // image and clip loader information files
  private static final String IMS_INFO = "imsInfo.txt";

  private static final int BLAST_FREQ = 5;  // used to judge when to fire a new blast


  private Thread animator;           // the thread that performs the animation
  private volatile boolean running = false;   // used to stop the animation thread
  private volatile boolean isPaused = false;

  private long period;                // period between drawing in _nanosecs_

  // off-screen rendering
  private Graphics2D dbg2; 
  private Image dbImage = null;

  private MagicWand gameTop;
  private BlastsManager blastsMan;     // manages the wand blasts
  private ImageAnalyzer imAnalyzer;    // analyzes the webcam images

  private BufferedImage bgImage = null;    // the background image
  private BufferedImage wandImage = null;  // a picture of the wand

  // data changed periodically by ImageAnalyzer
  private BufferedImage camImage = null;  // the current webcam image
  private Point wandHeadPt = null;        // the wand's head coordinates
  private double wandAngle = 0;           // angle made by wand to the horizontal

  private int blastCounter = 0;  
       // used to control the frequency of wand blasts

  

  public MWPanel(MagicWand mw, long period)
  {
    gameTop = mw;
    this.period = period;

    setDoubleBuffered(false);
    setBackground(Color.black);
    Dimension d = new Dimension(PWIDTH, PHEIGHT);
    setMinimumSize(d);
    setPreferredSize(d);

    setFocusable(true);
    requestFocus();    // the JPanel now has focus, so receives key events

	addKeyListener( new KeyAdapter() {
       public void keyPressed(KeyEvent e)
       { processKey(e);  }
     });

    // load the background and wand images
    ImagesLoader imsLoader = new ImagesLoader(IMS_INFO); 
    bgImage = imsLoader.getImage("desert");  // or try "arctic" or "space"
    wandImage = imsLoader.getImage("wand");

    // create manager for blast sprites
    blastsMan = new BlastsManager(PWIDTH, PHEIGHT, imsLoader, 
                                      (int)(period/1000000L) ); // in ms

    // start webcam snapping and analysis
    imAnalyzer = new ImageAnalyzer(this, PHEIGHT); 
         // the captured image is scaled to be the same height as the panel
    imAnalyzer.start();
  }  // end of MWPanel()



  private void processKey(KeyEvent e)
  // handles termination keys
  {
    int keyCode = e.getKeyCode();

    // termination keys
	// listen for esc, q, end, ctrl-c on the canvas to
	// allow a convenient exit from the full screen configuration
    if ((keyCode == KeyEvent.VK_ESCAPE) || (keyCode == KeyEvent.VK_Q) ||
        (keyCode == KeyEvent.VK_END) ||
        ((keyCode == KeyEvent.VK_C) && e.isControlDown()) )
      running = false;
  }  // end of processKey()



  public void addNotify()
  // wait for the JPanel to be added to the JFrame before starting
  { super.addNotify();   // creates the peer

    // initialise and start the game thread 
    if (animator == null || !running) {
      animator = new Thread(this);
	  animator.start();
    }
  } // end of addNotify()
    

  // ------------- game life cycle methods ------------
  // called by the JFrame's window listener methods


  public void resumeGame()
  // called when the JFrame is activated / deiconified
  {  isPaused = false;  } 


  public void pauseGame()
  // called when the JFrame is deactivated / iconified
  { isPaused = true;   } 


  public void stopGame() 
  // called when the JFrame is closing
  {  running = false; }

  // -----------------------------------------------

  public void run()
  /* The frames of the animation are drawn inside the while loop. */
  {
    long beforeTime, afterTime, timeDiff, sleepTime;
    long overSleepTime = 0L;
    int noDelays = 0;
    long excess = 0L;

    beforeTime = System.nanoTime();    // time at the start

	running = true;

	while(running) {
	  gameUpdate();
      gameRender();
      paintScreen();

      afterTime = System.nanoTime();
      timeDiff = afterTime - beforeTime;
      sleepTime = (period - timeDiff) - overSleepTime;  

      if (sleepTime > 0) {   // some time left in this cycle
        try {
          Thread.sleep(sleepTime/1000000L);  // nano -> ms
        }
        catch(InterruptedException ex){}
        overSleepTime = (System.nanoTime() - afterTime) - sleepTime;
      }
      else {    // sleepTime <= 0; the frame took longer than the period
        excess -= sleepTime;  // store excess time value
        overSleepTime = 0L;

        if (++noDelays >= NO_DELAYS_PER_YIELD) {
          Thread.yield();   // give another thread a chance to run
          noDelays = 0;
        }
      }

      beforeTime = System.nanoTime();

      /* If frame animation is taking too long, update the game state
         without rendering it, to get the updates/sec nearer to
         the required FPS. */
      int skips = 0;
      while((excess > period) && (skips < MAX_FRAME_SKIPS)) {
        excess -= period;
	    gameUpdate();    // update state but don't render
        skips++;
      }
	}

    imAnalyzer.closeDown();    // stop snapping pictures
    System.exit(0);   // so window disappears
  } // end of run()


  private void gameUpdate() 
  { if (!isPaused)
      blastsMan.updateSprites();  
  }  // end of gameUpdate()


  public void updateCamInfo(BufferedImage im, 
                                  Point hdPt, double angle)
  /* Called from ImageAnalyzer with the modified snap image, the
     coordinates of the wand head, and the wand's angle to the horizontal. 
     Since ImageAnalyzer is running in its own thread, this method
     may be called at any time. */
  {
    if (!isPaused) {
      camImage = im;
      wandHeadPt = hdPt;
      wandAngle = angle;
    }
  }  // end of updateCamInfo()


  // ------------------------------ rendering ---------------------


  private void gameRender()
  /* The game scene is composed from a background image, the
     webcam image, a picture of a wand, and blast sprites shot
     from the wand. */
  {
    if (dbImage == null){
      dbImage = createImage(PWIDTH, PHEIGHT);
      if (dbImage == null) {
        System.out.println("dbImage is null");
        return;
      }
      else
        dbg2 = (Graphics2D) dbImage.getGraphics();
    }

    // draw the background
    dbg2.drawImage(bgImage, 0, 0, this);

    if (camImage != null)  {   // if there's a webcam image
      int imXOffset = PWIDTH - camImage.getWidth();   
      /* this offset will put the right edge of webcam image 
         against the right edge of the panel */

      dbg2.drawImage(camImage, imXOffset, 0, this);   
         // draw the webcam image offset to the right

      if (wandHeadPt != null) {
        // copy wand head coords, and make the x-value relative to the panel
        int xHead = wandHeadPt.x + imXOffset;
        int yHead = wandHeadPt.y;
        double angle = wandAngle;   // copy the wand angle also

        drawWand(dbg2, xHead, yHead, angle); 
        fireWand(xHead, yHead, angle);
      }
    }
    blastsMan.drawSprites(dbg2);   // draw the blast sprites
  }  // end of gameRender()



  private void drawWand(Graphics2D g2, int xHead, int yHead, double angle)
  /* Draw the wand image after translating the drawing coordinates 
     system to the wand's head position, and rotating them by angle. 
     Use AffineTransform to modify the coordinates system. */ 
  {
    AffineTransform oldTrans = g2.getTransform(); // save current coords system

    AffineTransform posHead = new AffineTransform();
    posHead.translate(xHead-3, yHead-22);   
      /* Move the drawing coordinates to the head position.
         The (-3,-22) moves the coordinates back to the top-left of 
         the wand image. */

    // rotate the drawing coordinates by the wand angle amount
    posHead.rotate(angle); 

    g2.transform(posHead);  // apply the translation and rotation
    g2.drawImage(wandImage, 0, 0, this);    
        // draw the image at the modified (0,0) position

    // restore the old coordinate system
    g2.setTransform(oldTrans);
  }  // end of drawWand()


  private void fireWand(int xHead, int yHead, double angle)
  /* Request that the wand emit a blast sprite if the 
     blast counter is large enough. */
  {
    blastCounter++;
    if (blastCounter >= BLAST_FREQ) {
      blastsMan.requestBlast(xHead, yHead, angle);
      blastCounter = 0;   // reset counter
    }
  }  // end of fireWand()



  private void paintScreen()
  // use active rendering to put the buffered image on-screen
  { 
    Graphics g;
    try {
      g = this.getGraphics();
      if ((g != null) && (dbImage != null))
        g.drawImage(dbImage, 0, 0, null);
      // Sync the display on some systems.
      // (on Linux, this fixes event queue problems)
      Toolkit.getDefaultToolkit().sync();

      g.dispose();
    }
    catch (Exception e)
    { System.out.println("Graphics context error: " + e);  }
  } // end of paintScreen()


}  // end of MWPanel class
